# encoding: utf-8
# THIS FILE IS AUTOGENERATED!
from __future__ import unicode_literals
from setuptools import setup
setup(
    author='Kyle Lahnakoski',
    author_email='kyle@lahnakoski.com',
    classifiers=["Development Status :: 4 - Beta","Topic :: Software Development :: Libraries","Topic :: Software Development :: Libraries :: Python Modules","License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)","Programming Language :: Python :: 2.7"],
    description='JSON query expressions using Python',
    include_package_data=True,
    install_requires=["mo-collections>=3.77.20190","mo-dots>=3.77.20190","mo-files>=3.77.20190","mo-future>=3.71.20168","mo-json>=3.77.20190","mo-json-config>=3.77.20190","mo-kwargs>=3.77.20190","mo-logs>=3.75.20189","mo-math>=3.77.20190","mo-threads>=3.75.20189","mo-times>=3.76.20190"],
    license='MPL 2.0',
    long_description='# jx-python\nPython library for JSON Expressions \n',
    long_description_content_type='text/markdown',
    name='jx-python',
    packages=["jx_base","jx_base/expressions","jx_python/expressions","jx_python/containers","jx_python/cubes","jx_python/lists","jx_python/namespace","jx_python"],
    url='https://github.com/klahnakoski/jx-python',
    version='3.77.20190'
)